/*
 * dhcpcd - DHCP client daemon -
 * Copyright (C) 1996 - 1997 Yoichi Hariguchi <yoichi@fore.com>
 * Copyright (C) January, 1998 Sergei Viznyuk <sv@phystech.com>
 * 
 * dhcpcd is an RFC2131 and RFC1541 compliant DHCP client daemon.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Code Modified by Thomson. Revision 1.0.0
 */

#include <string.h>
#include <netinet/in.h>
#include <net/if_arp.h>
#include "client.h"
#include "udpipgen.h"

extern	dhcpMessage	*DhcpMsgSend;
extern	dhcpOptions	DhcpOptions;
extern  dhcpInterface   DhcpIface;
extern	char		*HostName;
extern	int		HostName_len;
extern	int		DebugFlag;
extern	int		BeRFC1541;
extern	unsigned	LeaseTime;
extern	int		TokenRingIf;
extern	unsigned char	ClientHwAddr[6];
extern  udpipMessage	UdpIpMsgSend;
extern  int 		magic_cookie;
extern  unsigned short  dhcpMsgSize;
extern  unsigned        nleaseTime;
extern  int             BroadcastResp;
extern  struct in_addr  inform_ipaddr;
extern	int		SendClientID;
extern	int		ask_subnetMask;
extern	int		ask_routersOnSubnet;
extern	int		ask_dns;
extern	int		ask_hostName;
extern	int		ask_domainName;
extern	int		ask_rootPath;
extern	int		ask_defaultIPTTL;
extern	int		ask_broadcastAddr;
extern	int		ask_performMaskDiscovery;
extern	int		ask_performRouterDiscovery;
extern	int		ask_staticRoute;
extern	int		ask_nisDomainName;
extern	int		ask_nisServers;
extern	int		ask_ntpServers;
extern	int		ask_dhcpIPaddrLeaseTime;
extern	int		ask_dhcpT1value;
extern	int		ask_dhcpT2value;
extern	int		ask_classlessStaticRoute;


/*****************************************************************************/
void buildDhcpDiscover(xid)
unsigned xid;
{
  register unsigned char *p = DhcpMsgSend->options + 4;
  register unsigned char *q = p;

/* build Ethernet header */
  memset(&UdpIpMsgSend,0,sizeof(udpipMessage));
  memcpy(UdpIpMsgSend.ethhdr.ether_dhost,MAC_BCAST_ADDR,ETH_ALEN);
  memcpy(UdpIpMsgSend.ethhdr.ether_shost,ClientHwAddr,ETH_ALEN);
  UdpIpMsgSend.ethhdr.ether_type = htons(ETHERTYPE_IP);

  DhcpMsgSend->op	=	DHCP_BOOTREQUEST;
  DhcpMsgSend->htype	=	(TokenRingIf) ? ARPHRD_IEEE802_TR : ARPHRD_ETHER;
  DhcpMsgSend->hlen	=	ETH_ALEN;
  DhcpMsgSend->xid	=	xid;
  DhcpMsgSend->secs	=       htons(10);
  if ( BroadcastResp )
    DhcpMsgSend->flags	=	htons(BROADCAST_FLAG);
#if 0
  memcpy(DhcpMsgSend->chaddr,DhcpIface.chaddr,ETH_ALEN);
#else
  memcpy(DhcpMsgSend->chaddr,ClientHwAddr,ETH_ALEN);
#endif
  memcpy(DhcpMsgSend->options,&magic_cookie,4);
  *p++ = dhcpMessageType;
  *p++ = 1;
  *p++ = DHCP_DISCOVER;
  *p++ = dhcpMaxMsgSize;
  *p++ = 2;
  memcpy(p,&dhcpMsgSize,2);
  p += 2;
  if ( DhcpIface.ciaddr )
    {
      if ( BeRFC1541 )
        DhcpMsgSend->ciaddr = DhcpIface.ciaddr;
      else
        {
          *p++ = dhcpRequestedIPaddr;
          *p++ = 4;
          memcpy(p,&DhcpIface.ciaddr,4);
          p += 4; 
        }
    }
  if (!ask_dhcpIPaddrLeaseTime) {
    *p++ = dhcpIPaddrLeaseTime;
    *p++ = 4;
    memcpy(p,&nleaseTime,4);
    p += 4;
  }
  *p++ = dhcpParamRequest;
  q = p;
  *p++ = 0;
  if (ask_subnetMask) { *p++ = subnetMask; (*q)++; }
  if (ask_routersOnSubnet) { *p++ = routersOnSubnet; (*q)++; }
  if (ask_dns) { *p++ = dns; (*q)++; }
  if (ask_hostName) { *p++ = hostName; (*q)++; }
  if (ask_domainName) { *p++ = domainName; (*q)++; }
  if (ask_rootPath) { *p++ = rootPath; (*q)++; }
  if (ask_defaultIPTTL) { *p++ = defaultIPTTL; (*q)++; }
  if (ask_broadcastAddr) { *p++ = broadcastAddr; (*q)++; }
  if (ask_performMaskDiscovery) { *p++ = performMaskDiscovery; (*q)++; }
  if (ask_performRouterDiscovery) { *p++ = performRouterDiscovery; (*q)++; }
  if (ask_staticRoute) { *p++ = staticRoute; (*q)++; }
  if (ask_nisDomainName) { *p++ = nisDomainName; (*q)++; }
  if (ask_nisServers) { *p++ = nisServers; (*q)++; }
  if (ask_ntpServers) { *p++ = ntpServers; (*q)++; }
  if (ask_dhcpIPaddrLeaseTime) { *p++ = dhcpIPaddrLeaseTime; (*q)++; }
  if (ask_dhcpT1value) { *p++ = dhcpT1value; (*q)++; }
  if (ask_dhcpT2value) { *p++ = dhcpT2value; (*q)++; }
  if (ask_classlessStaticRoute) { *p++ = classlessStaticRoute; (*q)++; }
  if ( HostName )
    {
      *p++ = hostName;
      *p++ = HostName_len;
      memcpy(p,HostName,HostName_len);
      p += HostName_len;
    }
  *p++ = dhcpClassIdentifier;
  *p++ = DhcpIface.class_len;
  memcpy(p,DhcpIface.class_id,DhcpIface.class_len);
  p += DhcpIface.class_len;
  if (SendClientID) {
    memcpy(p,DhcpIface.client_id,DhcpIface.client_len);
    p += DhcpIface.client_len;
  }
  *p = endOption;

/* build UDP/IP header */
  udpipgen((udpiphdr *)UdpIpMsgSend.udpipmsg,0,INADDR_BROADCAST,
  htons(DHCP_CLIENT_PORT),htons(DHCP_SERVER_PORT),sizeof(dhcpMessage));
}
/*****************************************************************************/
void buildDhcpRequest(xid)
unsigned xid;
{
  register unsigned char *p = DhcpMsgSend->options + 4;
  register unsigned char *q = p;
 
/* build Ethernet header */
  memset(&UdpIpMsgSend,0,sizeof(udpipMessage));
  memcpy(UdpIpMsgSend.ethhdr.ether_dhost,MAC_BCAST_ADDR,ETH_ALEN);
  memcpy(UdpIpMsgSend.ethhdr.ether_shost,ClientHwAddr,ETH_ALEN);
  UdpIpMsgSend.ethhdr.ether_type = htons(ETHERTYPE_IP);

  DhcpMsgSend->op	=	DHCP_BOOTREQUEST;
  DhcpMsgSend->htype	=	(TokenRingIf) ? ARPHRD_IEEE802_TR : ARPHRD_ETHER;
  DhcpMsgSend->hlen	=	ETH_ALEN;
  DhcpMsgSend->xid	=	xid;
  DhcpMsgSend->secs	=	htons(10);
  if ( BroadcastResp )
    DhcpMsgSend->flags	=	htons(BROADCAST_FLAG);
#if 0
  memcpy(DhcpMsgSend->chaddr,DhcpIface.chaddr,ETH_ALEN);
#else
  memcpy(DhcpMsgSend->chaddr,ClientHwAddr,ETH_ALEN);
#endif
  memcpy(DhcpMsgSend->options,&magic_cookie,4);
  *p++ = dhcpMessageType;
  *p++ = 1;
  *p++ = DHCP_REQUEST;
  *p++ = dhcpMaxMsgSize;
  *p++ = 2;
  memcpy(p,&dhcpMsgSize,2);
  p += 2;
  *p++ = dhcpServerIdentifier;
  *p++ = 4;
  memcpy(p,DhcpOptions.val[dhcpServerIdentifier],4);
  p += 4;
  if ( BeRFC1541 )
    DhcpMsgSend->ciaddr = DhcpIface.ciaddr;
  else
    {
      *p++ = dhcpRequestedIPaddr;
      *p++ = 4;
      memcpy(p,&DhcpIface.ciaddr,4);
      p += 4;
    }
  if ( DhcpOptions.val[dhcpIPaddrLeaseTime] )
    {
      *p++ = dhcpIPaddrLeaseTime;
      *p++ = 4;
      memcpy(p,DhcpOptions.val[dhcpIPaddrLeaseTime],4);
      p += 4;
    }
  *p++ = dhcpParamRequest;
  q = p;
  *p++ = 0;
  if (ask_subnetMask) { *p++ = subnetMask; (*q)++; }
  if (ask_routersOnSubnet) { *p++ = routersOnSubnet; (*q)++; }
  if (ask_dns) { *p++ = dns; (*q)++; }
  if (ask_hostName) { *p++ = hostName; (*q)++; }
  if (ask_domainName) { *p++ = domainName; (*q)++; }
  if (ask_rootPath) { *p++ = rootPath; (*q)++; }
  if (ask_defaultIPTTL) { *p++ = defaultIPTTL; (*q)++; }
  if (ask_broadcastAddr) { *p++ = broadcastAddr; (*q)++; }
  if (ask_performMaskDiscovery) { *p++ = performMaskDiscovery; (*q)++; }
  if (ask_performRouterDiscovery) { *p++ = performRouterDiscovery; (*q)++; }
  if (ask_staticRoute) { *p++ = staticRoute; (*q)++; }
  if (ask_nisDomainName) { *p++ = nisDomainName; (*q)++; }
  if (ask_nisServers) { *p++ = nisServers; (*q)++; }
  if (ask_ntpServers) { *p++ = ntpServers; (*q)++; }
  if (ask_dhcpIPaddrLeaseTime && !DhcpOptions.val[dhcpIPaddrLeaseTime]) { *p++ = dhcpIPaddrLeaseTime; (*q)++; }
  if (ask_dhcpT1value) { *p++ = dhcpT1value; (*q)++; }
  if (ask_dhcpT2value) { *p++ = dhcpT2value; (*q)++; }
  if (ask_classlessStaticRoute) { *p++ = classlessStaticRoute; (*q)++; }
  if ( HostName )
    {
      *p++ = hostName;
      *p++ = HostName_len;
      memcpy(p,HostName,HostName_len);
      p += HostName_len;
    }
  *p++ = dhcpClassIdentifier;
  *p++ = DhcpIface.class_len;
  memcpy(p,DhcpIface.class_id,DhcpIface.class_len);
  p += DhcpIface.class_len;
  if (SendClientID) {
    memcpy(p,DhcpIface.client_id,DhcpIface.client_len);
    p += DhcpIface.client_len;
  }
  *p = endOption;

/* build UDP/IP header */
  udpipgen((udpiphdr *)UdpIpMsgSend.udpipmsg,0,INADDR_BROADCAST,
  htons(DHCP_CLIENT_PORT),htons(DHCP_SERVER_PORT),sizeof(dhcpMessage));
}
/*****************************************************************************/
void buildDhcpRenew(xid)
unsigned xid;
{
  register unsigned char *p = DhcpMsgSend->options + 4;
  register unsigned char *q = p;
  memset(&UdpIpMsgSend,0,sizeof(udpipMessage));
  memcpy(UdpIpMsgSend.ethhdr.ether_dhost,DhcpIface.shaddr,ETH_ALEN);
  memcpy(UdpIpMsgSend.ethhdr.ether_shost,ClientHwAddr,ETH_ALEN);
  UdpIpMsgSend.ethhdr.ether_type = htons(ETHERTYPE_IP);

  DhcpMsgSend->op	=	DHCP_BOOTREQUEST;
  DhcpMsgSend->htype	=	(TokenRingIf) ? ARPHRD_IEEE802_TR : ARPHRD_ETHER;
  DhcpMsgSend->hlen	=	ETH_ALEN;
  DhcpMsgSend->xid	=	xid;
  DhcpMsgSend->secs	=	htons(10);
  if ( BroadcastResp )
    DhcpMsgSend->flags	=	htons(BROADCAST_FLAG);
  DhcpMsgSend->ciaddr   =       DhcpIface.ciaddr;
#if 0
  memcpy(DhcpMsgSend->chaddr,DhcpIface.chaddr,ETH_ALEN);
#else
  memcpy(DhcpMsgSend->chaddr,ClientHwAddr,ETH_ALEN);
#endif
  memcpy(DhcpMsgSend->options,&magic_cookie,4);
  *p++ = dhcpMessageType;
  *p++ = 1;
  *p++ = DHCP_REQUEST;
  *p++ = dhcpMaxMsgSize;
  *p++ = 2;
  memcpy(p,&dhcpMsgSize,2);
  p += 2;
#if 0
  if ( DhcpOptions.val[dhcpIPaddrLeaseTime] )
    {
      *p++ = dhcpIPaddrLeaseTime;
      *p++ = 4;
      memcpy(p,DhcpOptions.val[dhcpIPaddrLeaseTime],4);
      p += 4;
    }
#endif
  *p++ = dhcpParamRequest;
  q = p;
  *p++ = 0;
  if (ask_subnetMask) { *p++ = subnetMask; (*q)++; }
  if (ask_routersOnSubnet) { *p++ = routersOnSubnet; (*q)++; }
  if (ask_dns) { *p++ = dns; (*q)++; }
  if (ask_hostName) { *p++ = hostName; (*q)++; }
  if (ask_domainName) { *p++ = domainName; (*q)++; }
  if (ask_rootPath) { *p++ = rootPath; (*q)++; }
  if (ask_defaultIPTTL) { *p++ = defaultIPTTL; (*q)++; }
  if (ask_broadcastAddr) { *p++ = broadcastAddr; (*q)++; }
  if (ask_performMaskDiscovery) { *p++ = performMaskDiscovery; (*q)++; }
  if (ask_performRouterDiscovery) { *p++ = performRouterDiscovery; (*q)++; }
  if (ask_staticRoute) { *p++ = staticRoute; (*q)++; }
  if (ask_nisDomainName) { *p++ = nisDomainName; (*q)++; }
  if (ask_nisServers) { *p++ = nisServers; (*q)++; }
  if (ask_ntpServers) { *p++ = ntpServers; (*q)++; }
  if (ask_dhcpIPaddrLeaseTime) { *p++ = dhcpIPaddrLeaseTime; (*q)++; }
  if (ask_dhcpT1value) { *p++ = dhcpT1value; (*q)++; }
  if (ask_dhcpT2value) { *p++ = dhcpT2value; (*q)++; }
  if (ask_classlessStaticRoute) { *p++ = classlessStaticRoute; (*q)++; }
  if ( HostName )
    {
      *p++ = hostName;
      *p++ = HostName_len;
      memcpy(p,HostName,HostName_len);
      p += HostName_len;
    }
  *p++ = dhcpClassIdentifier;
  *p++ = DhcpIface.class_len;
  memcpy(p,DhcpIface.class_id,DhcpIface.class_len);
  p += DhcpIface.class_len;
  if (SendClientID) {
    memcpy(p,DhcpIface.client_id,DhcpIface.client_len);
    p += DhcpIface.client_len;
  }
  *p = endOption;

  udpipgen((udpiphdr *)UdpIpMsgSend.udpipmsg,
  DhcpIface.ciaddr,DhcpIface.siaddr,
  htons(DHCP_CLIENT_PORT),htons(DHCP_SERVER_PORT),sizeof(dhcpMessage));
}
/*****************************************************************************/
void buildDhcpRebind(xid)
unsigned xid;
{
  register unsigned char *p = DhcpMsgSend->options + 4;
  register unsigned char *q = p;
  memset(&UdpIpMsgSend,0,sizeof(udpipMessage));
  memcpy(UdpIpMsgSend.ethhdr.ether_dhost,MAC_BCAST_ADDR,ETH_ALEN);
  memcpy(UdpIpMsgSend.ethhdr.ether_shost,ClientHwAddr,ETH_ALEN);
  UdpIpMsgSend.ethhdr.ether_type = htons(ETHERTYPE_IP);

  DhcpMsgSend->op	=	DHCP_BOOTREQUEST;
  DhcpMsgSend->htype	=	(TokenRingIf) ? ARPHRD_IEEE802_TR : ARPHRD_ETHER;
  DhcpMsgSend->hlen	=	ETH_ALEN;
  DhcpMsgSend->xid	=	xid;
  DhcpMsgSend->secs	=	htons(10);
  if ( BroadcastResp )
    DhcpMsgSend->flags	=	htons(BROADCAST_FLAG);
  DhcpMsgSend->ciaddr   =       DhcpIface.ciaddr;
#if 0
  memcpy(DhcpMsgSend->chaddr,DhcpIface.chaddr,ETH_ALEN);
#else
  memcpy(DhcpMsgSend->chaddr,ClientHwAddr,ETH_ALEN);
#endif
  memcpy(DhcpMsgSend->options,&magic_cookie,4);
  *p++ = dhcpMessageType;
  *p++ = 1;
  *p++ = DHCP_REQUEST;
  *p++ = dhcpMaxMsgSize;
  *p++ = 2;
  memcpy(p,&dhcpMsgSize,2);
  p += 2;
  if ( DhcpOptions.val[dhcpIPaddrLeaseTime] )
    {
      *p++ = dhcpIPaddrLeaseTime;
      *p++ = 4;
      memcpy(p,DhcpOptions.val[dhcpIPaddrLeaseTime],4);
      p += 4;
    }
  *p++ = dhcpParamRequest;
  q = p;
  *p++ = 0;
  if (ask_subnetMask) { *p++ = subnetMask; (*q)++; }
  if (ask_routersOnSubnet) { *p++ = routersOnSubnet; (*q)++; }
  if (ask_dns) { *p++ = dns; (*q)++; }
  if (ask_hostName) { *p++ = hostName; (*q)++; }
  if (ask_domainName) { *p++ = domainName; (*q)++; }
  if (ask_rootPath) { *p++ = rootPath; (*q)++; }
  if (ask_defaultIPTTL) { *p++ = defaultIPTTL; (*q)++; }
  if (ask_broadcastAddr) { *p++ = broadcastAddr; (*q)++; }
  if (ask_performMaskDiscovery) { *p++ = performMaskDiscovery; (*q)++; }
  if (ask_performRouterDiscovery) { *p++ = performRouterDiscovery; (*q)++; }
  if (ask_staticRoute) { *p++ = staticRoute; (*q)++; }
  if (ask_nisDomainName) { *p++ = nisDomainName; (*q)++; }
  if (ask_nisServers) { *p++ = nisServers; (*q)++; }
  if (ask_ntpServers) { *p++ = ntpServers; (*q)++; }
  if (ask_dhcpIPaddrLeaseTime && !DhcpOptions.val[dhcpIPaddrLeaseTime]) { *p++ = dhcpIPaddrLeaseTime; (*q)++; }
  if (ask_dhcpT1value) { *p++ = dhcpT1value; (*q)++; }
  if (ask_dhcpT2value) { *p++ = dhcpT2value; (*q)++; }
  if (ask_classlessStaticRoute) { *p++ = classlessStaticRoute; (*q)++; }
  if ( HostName )
    {
      *p++ = hostName;
      *p++ = HostName_len;
      memcpy(p,HostName,HostName_len);
      p += HostName_len;
    }
  *p++ = dhcpClassIdentifier;
  *p++ = DhcpIface.class_len;
  memcpy(p,DhcpIface.class_id,DhcpIface.class_len);
  p += DhcpIface.class_len;
  if (SendClientID) {
    memcpy(p,DhcpIface.client_id,DhcpIface.client_len);
    p += DhcpIface.client_len;
  }
  *p = endOption;

  udpipgen((udpiphdr *)UdpIpMsgSend.udpipmsg,
  DhcpIface.ciaddr,INADDR_BROADCAST,
  htons(DHCP_CLIENT_PORT),htons(DHCP_SERVER_PORT),sizeof(dhcpMessage));
}
/*****************************************************************************/
void buildDhcpReboot(xid)
unsigned xid;
{
  register unsigned char *p = DhcpMsgSend->options + 4;
  register unsigned char *q = p;
 
/* build Ethernet header */
  memset(&UdpIpMsgSend,0,sizeof(udpipMessage));
  memcpy(UdpIpMsgSend.ethhdr.ether_dhost,MAC_BCAST_ADDR,ETH_ALEN);
  memcpy(UdpIpMsgSend.ethhdr.ether_shost,ClientHwAddr,ETH_ALEN);
  UdpIpMsgSend.ethhdr.ether_type = htons(ETHERTYPE_IP);

  DhcpMsgSend->op	=	DHCP_BOOTREQUEST;
  DhcpMsgSend->htype	=	(TokenRingIf) ? ARPHRD_IEEE802_TR : ARPHRD_ETHER;
  DhcpMsgSend->hlen	=	ETH_ALEN;
  DhcpMsgSend->xid	=	xid;
  DhcpMsgSend->secs	=	htons(10);
  if ( BroadcastResp )
    DhcpMsgSend->flags	=	htons(BROADCAST_FLAG);
#if 0
  memcpy(DhcpMsgSend->chaddr,DhcpIface.chaddr,ETH_ALEN);
#else
  memcpy(DhcpMsgSend->chaddr,ClientHwAddr,ETH_ALEN);
#endif
  memcpy(DhcpMsgSend->options,&magic_cookie,4);

  *p++ = dhcpMessageType;
  *p++ = 1;
  *p++ = DHCP_REQUEST;
  *p++ = dhcpMaxMsgSize;
  *p++ = 2;
  memcpy(p,&dhcpMsgSize,2);
  p += 2;
  if ( BeRFC1541 )
    DhcpMsgSend->ciaddr = DhcpIface.ciaddr;
  else
    {
      *p++ = dhcpRequestedIPaddr;
      *p++ = 4;
      memcpy(p,&DhcpIface.ciaddr,4);
      p += 4;
    }
  if (!ask_dhcpIPaddrLeaseTime) {
    *p++ = dhcpIPaddrLeaseTime;
    *p++ = 4;
    memcpy(p,&nleaseTime,4);
    p += 4;
  }
  *p++ = dhcpParamRequest;
  q = p;
  *p++ = 0;
  if (ask_subnetMask) { *p++ = subnetMask; (*q)++; }
  if (ask_routersOnSubnet) { *p++ = routersOnSubnet; (*q)++; }
  if (ask_dns) { *p++ = dns; (*q)++; }
  if (ask_hostName) { *p++ = hostName; (*q)++; }
  if (ask_domainName) { *p++ = domainName; (*q)++; }
  if (ask_rootPath) { *p++ = rootPath; (*q)++; }
  if (ask_defaultIPTTL) { *p++ = defaultIPTTL; (*q)++; }
  if (ask_broadcastAddr) { *p++ = broadcastAddr; (*q)++; }
  if (ask_performMaskDiscovery) { *p++ = performMaskDiscovery; (*q)++; }
  if (ask_performRouterDiscovery) { *p++ = performRouterDiscovery; (*q)++; }
  if (ask_staticRoute) { *p++ = staticRoute; (*q)++; }
  if (ask_nisDomainName) { *p++ = nisDomainName; (*q)++; }
  if (ask_nisServers) { *p++ = nisServers; (*q)++; }
  if (ask_ntpServers) { *p++ = ntpServers; (*q)++; }
  if (ask_dhcpIPaddrLeaseTime) { *p++ = dhcpIPaddrLeaseTime; (*q)++; }
  if (ask_dhcpT1value) { *p++ = dhcpT1value; (*q)++; }
  if (ask_dhcpT2value) { *p++ = dhcpT2value; (*q)++; }
  if (ask_classlessStaticRoute) { *p++ = classlessStaticRoute; (*q)++; }
  if ( HostName )
    {
      *p++ = hostName;
      *p++ = HostName_len;
      memcpy(p,HostName,HostName_len);
      p += HostName_len;
    }
  *p++ = dhcpClassIdentifier;
  *p++ = DhcpIface.class_len;
  memcpy(p,DhcpIface.class_id,DhcpIface.class_len);
  p += DhcpIface.class_len;
  if (SendClientID) {
    memcpy(p,DhcpIface.client_id,DhcpIface.client_len);
    p += DhcpIface.client_len;
  }
  *p = endOption;

  udpipgen((udpiphdr *)UdpIpMsgSend.udpipmsg,0,INADDR_BROADCAST,
  htons(DHCP_CLIENT_PORT),htons(DHCP_SERVER_PORT),sizeof(dhcpMessage));
}
/*****************************************************************************/
void buildDhcpRelease(xid)
unsigned xid;
{
  register unsigned char *p = DhcpMsgSend->options + 4;
  memset(&UdpIpMsgSend,0,sizeof(udpipMessage));
  memcpy(UdpIpMsgSend.ethhdr.ether_dhost,DhcpIface.shaddr,ETH_ALEN);
  memcpy(UdpIpMsgSend.ethhdr.ether_shost,ClientHwAddr,ETH_ALEN);
  UdpIpMsgSend.ethhdr.ether_type = htons(ETHERTYPE_IP);

  DhcpMsgSend->op	=	DHCP_BOOTREQUEST;
  DhcpMsgSend->htype	=	(TokenRingIf) ? ARPHRD_IEEE802_TR : ARPHRD_ETHER;
  DhcpMsgSend->hlen	=	ETH_ALEN;
  DhcpMsgSend->xid	=	xid;
  DhcpMsgSend->ciaddr	=	DhcpIface.ciaddr;
#if 0
  memcpy(DhcpMsgSend->chaddr,DhcpIface.chaddr,ETH_ALEN);
#else
  memcpy(DhcpMsgSend->chaddr,ClientHwAddr,ETH_ALEN);
#endif
  memcpy(DhcpMsgSend->options,&magic_cookie,4);
  *p++ = dhcpMessageType;
  *p++ = 1;
  *p++ = DHCP_RELEASE;
  *p++ = dhcpServerIdentifier;
  *p++ = 4;
  memcpy(p,DhcpOptions.val[dhcpServerIdentifier],4);
  p += 4;
  if (SendClientID) {
    memcpy(p,DhcpIface.client_id,DhcpIface.client_len);
    p += DhcpIface.client_len;
  }
  *p = endOption;

  udpipgen((udpiphdr *)UdpIpMsgSend.udpipmsg,DhcpIface.ciaddr,
  DhcpIface.siaddr,htons(DHCP_CLIENT_PORT),htons(DHCP_SERVER_PORT),
  sizeof(dhcpMessage));
}
/*****************************************************************************/
#ifdef ARPCHECK
void buildDhcpDecline(xid)
unsigned xid;
{
  register unsigned char *p = DhcpMsgSend->options + 4;
  memset(&UdpIpMsgSend,0,sizeof(udpipMessage));
  memcpy(UdpIpMsgSend.ethhdr.ether_dhost,DhcpIface.shaddr,ETH_ALEN);
  memcpy(UdpIpMsgSend.ethhdr.ether_shost,ClientHwAddr,ETH_ALEN);
  UdpIpMsgSend.ethhdr.ether_type = htons(ETHERTYPE_IP);

  DhcpMsgSend->op	=	DHCP_BOOTREQUEST;
  DhcpMsgSend->htype	=	(TokenRingIf) ? ARPHRD_IEEE802_TR : ARPHRD_ETHER;
  DhcpMsgSend->hlen	=	ETH_ALEN;
  DhcpMsgSend->xid	=	xid;
#if 0
  memcpy(DhcpMsgSend->chaddr,DhcpIface.chaddr,ETH_ALEN);
#else
  memcpy(DhcpMsgSend->chaddr,ClientHwAddr,ETH_ALEN);
#endif
  memcpy(DhcpMsgSend->options,&magic_cookie,4);
  *p++ = dhcpMessageType;
  *p++ = 1;
  *p++ = DHCP_DECLINE;
  *p++ = dhcpServerIdentifier;
  *p++ = 4;
  memcpy(p,DhcpOptions.val[dhcpServerIdentifier],4);
  p += 4;
  if ( BeRFC1541 )
    DhcpMsgSend->ciaddr = DhcpIface.ciaddr;
  else
    {
      *p++ = dhcpRequestedIPaddr;
      *p++ = 4;
      memcpy(p,&DhcpIface.ciaddr,4);
      p += 4;
    }
  if (SendClientID) {
    memcpy(p,DhcpIface.client_id,DhcpIface.client_len);
    p += DhcpIface.client_len;
  }
  *p = endOption;

  udpipgen((udpiphdr *)UdpIpMsgSend.udpipmsg,0,
  DhcpIface.siaddr,htons(DHCP_CLIENT_PORT),htons(DHCP_SERVER_PORT),
  sizeof(dhcpMessage));
}
#endif
/*****************************************************************************/
void buildDhcpInform(xid)
unsigned xid;
{
  register unsigned char *p = DhcpMsgSend->options + 4;
  register unsigned char *q = p;

  memset(&UdpIpMsgSend,0,sizeof(udpipMessage));
  memcpy(UdpIpMsgSend.ethhdr.ether_dhost,MAC_BCAST_ADDR,ETH_ALEN);
  memcpy(UdpIpMsgSend.ethhdr.ether_shost,ClientHwAddr,ETH_ALEN);
  UdpIpMsgSend.ethhdr.ether_type = htons(ETHERTYPE_IP);

  DhcpMsgSend->op	=	DHCP_BOOTREQUEST;
  DhcpMsgSend->htype	=	(TokenRingIf) ? ARPHRD_IEEE802_TR : ARPHRD_ETHER;
  DhcpMsgSend->hlen	=	ETH_ALEN;
  DhcpMsgSend->xid	=	xid;
  DhcpMsgSend->secs	=       htons(10);
  if ( BroadcastResp )
    DhcpMsgSend->flags	=	htons(BROADCAST_FLAG);
#if 0
  memcpy(DhcpMsgSend->chaddr,DhcpIface.chaddr,ETH_ALEN);
#else
  memcpy(DhcpMsgSend->chaddr,ClientHwAddr,ETH_ALEN);
#endif
  DhcpMsgSend->ciaddr = inform_ipaddr.s_addr;
  memcpy(DhcpMsgSend->options,&magic_cookie,4);
  *p++ = dhcpMessageType;
  *p++ = 1;
  *p++ = DHCP_INFORM;
  *p++ = dhcpMaxMsgSize;
  *p++ = 2;
  memcpy(p,&dhcpMsgSize,2);
  p += 2;
  *p++ = dhcpParamRequest;
  q = p;
  *p++ = 0;
  if (ask_subnetMask) { *p++ = subnetMask; (*q)++; }
  if (ask_routersOnSubnet) { *p++ = routersOnSubnet; (*q)++; }
  if (ask_dns) { *p++ = dns; (*q)++; }
  if (ask_hostName) { *p++ = hostName; (*q)++; }
  if (ask_domainName) { *p++ = domainName; (*q)++; }
  if (ask_rootPath) { *p++ = rootPath; (*q)++; }
  if (ask_defaultIPTTL) { *p++ = defaultIPTTL; (*q)++; }
  if (ask_broadcastAddr) { *p++ = broadcastAddr; (*q)++; }
  if (ask_performMaskDiscovery) { *p++ = performMaskDiscovery; (*q)++; }
  if (ask_performRouterDiscovery) { *p++ = performRouterDiscovery; (*q)++; }
  if (ask_staticRoute) { *p++ = staticRoute; (*q)++; }
  if (ask_nisDomainName) { *p++ = nisDomainName; (*q)++; }
  if (ask_nisServers) { *p++ = nisServers; (*q)++; }
  if (ask_ntpServers) { *p++ = ntpServers; (*q)++; }
  if (ask_dhcpIPaddrLeaseTime) { *p++ = dhcpIPaddrLeaseTime; (*q)++; }
  if (ask_dhcpT1value) { *p++ = dhcpT1value; (*q)++; }
  if (ask_dhcpT2value) { *p++ = dhcpT2value; (*q)++; }
  if (ask_classlessStaticRoute) { *p++ = classlessStaticRoute; (*q)++; }
  if ( HostName )
    {
      *p++ = hostName;
      *p++ = HostName_len;
      memcpy(p,HostName,HostName_len);
      p += HostName_len;
    }
  *p++ = dhcpClassIdentifier;
  *p++ = DhcpIface.class_len;
  memcpy(p,DhcpIface.class_id,DhcpIface.class_len);
  p += DhcpIface.class_len;
  if (SendClientID) {
    memcpy(p,DhcpIface.client_id,DhcpIface.client_len);
    p += DhcpIface.client_len;
  }
  *p = endOption;

  udpipgen((udpiphdr *)UdpIpMsgSend.udpipmsg,0,INADDR_BROADCAST,
  htons(DHCP_CLIENT_PORT),htons(DHCP_SERVER_PORT),sizeof(dhcpMessage));
}
